@@ -235,7 +235,8 @@ void do_exec(void)
   if (!cvm_setugid())
     respond(0, "NO Internal error: could not set UID/GID");
   else if (!cvm_setenv() ||
-	   setenv("MAILDIR", cvm_fact_mailbox, 1) == -1 ||
+	   (cvm_fact_mailbox != 0 &&
+	    setenv("MAILDIR", cvm_fact_mailbox, 1) == -1) ||
 	   setenv("IMAPLOGINTAG", tag.s, 1) == -1 ||
 	   setenv("AUTHENTICATED", cvm_fact_username, 1) == -1)
     respond(0, "NO Internal error: could not set environment");
