@@ -6,7 +6,9 @@ void insthier(void) {
   c(bin, "imapfront-auth",    -1, -1, 0755);
   c(bin, "pop3front-auth",    -1, -1, 0755);
   c(bin, "pop3front-maildir", -1, -1, 0755);
+  c(bin, "qmqpfront-echo",    -1, -1, 0755);
   c(bin, "qmqpfront-qmail",   -1, -1, 0755);
+  c(bin, "qmtpfront-echo",    -1, -1, 0755);
   c(bin, "qmtpfront-qmail",   -1, -1, 0755);
   c(bin, "smtpfront-echo",    -1, -1, 0755);
   c(bin, "smtpfront-qmail",   -1, -1, 0755);
